"""
data loading and transforming pandas dataframe to Tensorflow Dataset
"""

import pandas
import tensorflow as tf


class TensorflowDataMapper:
    """Transform pandas data frame to tensorflow data set
    """
    def __init__(self):
        pass

    def data_frame_to_dataset(self, data_frame_features=None, data_frame_labels=None):
        """Transform from data frame to data_set

        Args:
            data_frame_features (pandas.DataFrame): Features Data.
            data_frame_labels (pandas.DataFrame): Target Data
        """
        if data_frame_labels is not None:
            return tf.data.Dataset.from_tensor_slices((dict(data_frame_features), data_frame_labels.values))
        return tf.data.Dataset.from_tensor_slices(dict(data_frame_features))

    def map(self, data_frame_features=None, data_frame_labels=None):
        """map pandas dataframe to tf.data.Dataset

        Args:
            data_frame_features (pandas.DataFrame): Features Data.
            data_frame_labels (pandas.DataFrame): Target Data
        """
        return self.data_frame_to_dataset(data_frame_features, data_frame_labels)

    def split_data_set(self, dataset=None, val_split_fraction=0.25):
        """Split data for training and validation

        Args:
            features (tf.data.Dataset): Dataset to split into training and validation
        
        Returns:
            (tf.data.Dataset, tf.data.Dataset): train and validation datasets
        """
        rows = dataset.cardinality().numpy()
        dataset = dataset.shuffle(rows)
        training_size = int((1 - val_split_fraction) * rows)
        train_data_set = dataset.take(training_size)
        val_data_set = dataset.skip(training_size)
        return train_data_set, val_data_set



import unittest
import pandas as pd
import tensorflow as tf

# local imports
from easyflow.data import TensorflowDataMapper
from easyflow.feature_encoders import FeatureUnionTransformer
from easyflow.feature_encoders import NumericalFeatureEncoder, CategoricalFeatureEncoder, BucketizedFeatureEncoder,\
    EmbeddingFeatureEncoder, CategoryCrossingFeatureEncoder


class TestFeatureEncoders(unittest.TestCase):
    """Test feature encoder module
    """
    def setUp(self):
        """Setup test data and pipeline objects
        """
        dataframe = pd.read_csv('easyflow/tests/test_data/heart.csv')
        dataframe['age'] *= 1.0
        labels = dataframe.pop("target")
        dataset_mapper = TensorflowDataMapper()
        self.dataset = dataset_mapper.map(dataframe, labels).batch(32)

        self.numerical_features = ['trestbps', 'chol', 'thalach', 'oldpeak', 'slope']
        self.categorical_features = ['sex', 'cp', 'fbs', 'restecg', 'exang', 'ca', 'thal']

    def test_feature_union(self):
        """Test the feature encoder pipeline by applying a full model
        """
        # Below encoder list is only for testing/illustration purposes
        feature_encoder_list = [('numerical_features', NumericalFeatureEncoder(), self.numerical_features),
                                ('categorical_features', CategoricalFeatureEncoder(), self.categorical_features),
                                ('embedding_features', EmbeddingFeatureEncoder(dimension=3), ['thal']),
                                ('bucketized_age', BucketizedFeatureEncoder(boundaries=[20, 50, 90]), ['age']),
                                ('cross_column', CategoryCrossingFeatureEncoder(hash_bucket_size=10), ['sex', 'thal'])
                                ]

        feature_layer_inputs, feature_layer =  FeatureUnionTransformer(feature_encoder_list).transform(self.dataset)

        x = tf.keras.layers.BatchNormalization()(feature_layer)
        x = tf.keras.layers.Dense(128, activation="relu")(x)
        x = tf.keras.layers.Dropout(0.5)(x)
        outputs = tf.keras.layers.Dense(1, activation='sigmoid')(x)
        model = tf.keras.Model(inputs=[v for v in feature_layer_inputs.values()], outputs=outputs)
        model.compile(optimizer=tf.keras.optimizers.Adam(),
                      loss=tf.keras.losses.BinaryCrossentropy(),
                      metrics=[tf.keras.metrics.BinaryAccuracy(name='accuracy'), tf.keras.metrics.AUC(name='auc')])
        history=model.fit(self.dataset, epochs=10)

        # test if the model ran through all 10 epochs
        assert len(history.history['loss']) == 10
        assert len(feature_layer_inputs) == 13
        assert feature_layer.shape[1] == 45

    def test_infered_encoding(self):
        """Test the basic infered encoding transformer"""
        encoder = FeatureUnionTransformer.infer_feature_transformer(self.dataset)
        feature_layer_inputs, feature_layer = encoder.transform(self.dataset)
        # user should ensure dtypes are correct
        assert feature_layer.shape[-1] == 321


if __name__ == '__main__':
    unittest.main()



"""Main feature preprocessing interfaces"""

import tensorflow as tf
from easyflow.preprocessing.factory import (
    FeaturePreprocessorFactory,
)


class FeaturePreprocessor(FeaturePreprocessorFactory):
    """
    Main interface for transforming features. Apply feature preprocessing list which can contain both
    Encoder and SequentialEncoder object types
    Args:
        feature_encoder_list : List of encoders of the form: ('name', encoder type, list of features)
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def call(self, inputs):
        """Join preprocessing steps.

        Args:
            dataset ([tf.data.Dataset, pandas.DataFrame]): Features Data to apply encoder on.

        Returns:
            (Bunch): Container object that holds feature preprocessing layer steps.
        """
        preprocessed_input = super(FeaturePreprocessor, self).call(inputs)
        # Should this concatenation happen in feature_preprocessor_layer call methods?
        preprocessed_input = {
            step_name: tf.keras.layers.concatenate(step_layers)
            for step_name, step_layers in preprocessed_input.items()
        }
        return Bunch(**preprocessed_input)


class FeatureUnion(FeaturePreprocessorFactory):
    """Apply column based preprocessing on the data and combine features with a concat layer.

    Args:
        feature_encoder_list : List of encoders of the form: ('name', encoder type, list of features)
    """

    def __init__(self, *args, **kwargs):
        super(FeatureUnion, self).__init__(*args, **kwargs)

    def call(self, inputs):
        """Join features. If more flexibility and customization is needed use PreprocessorColumnTransformer.

        Args:
            dataset (tf.data.Dataset): Features Data to apply encoder on.

        Returns:
            (list, tf.keras.layer): Keras inputs for each feature and concatenated layer
        """
        preprocessed_input = super(FeatureUnion, self).call(inputs)
        # flatten (or taking the union) of preprocessed inputs
        preprocessed_input = [
            low_level
            for _, high_level in preprocessed_input.items()
            for low_level in high_level
        ]
        if len(preprocessed_input) > 1:
            return tf.keras.layers.concatenate(preprocessed_input)
        return preprocessed_input.pop()


class Bunch(dict):
    """Helper class that bunch set of layer steps into one and can easily be 
    accessed as preprocessor.step_name

    Reference: https://github.com/scikit-learn/scikit-learn/sklearn/utils/_bunch.py
    """
    def __init__(self, **kwargs):
        super().__init__(kwargs)

    def __setattr__(self, key, value):
        self[key] = value

    def __dir__(self):
        return self.keys()

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key)



"""base classes for stateful preprocessing layers"""
import tensorflow as tf
from tensorflow.keras import layers

from easyflow.preprocessing.custom import StringToIntegerLookup
from easyflow.preprocessing.base import BaseFeaturePreprocessor


class FeaturePreprocessorFromTensorflowDataset(tf.keras.layers.Layer, BaseFeaturePreprocessor):
    """Feature Preprocessing Layer for Tensorflow Dataset type. The class takes in steps of preprocessing that
    needs to be applied on the dataset. The adapt method records all stateful parameters for each of the steps
    and features that will be adapted. These adapted layers will be stored as an attribute and applied in the
    forward pass of the network.

    Args:
        feature_encoder_list : List of preprocessor of the form: ('name', preprocessor type, list of features)

    """

    def __init__(self, feature_preprocessor_list=[], *args, **kwargs):
        super(FeaturePreprocessorFromTensorflowDataset, self).__init__(*args, *kwargs)
        self.feature_preprocessor_list = feature_preprocessor_list
        self.feature_preprocessor_list = self.map_preprocessor(self.feature_preprocessor_list)
        self.adapted_preprocessors = dict()

    def adapt(self, dataset, *args, **kwargs):
        """Loop through the feature preprocessor list and dapt features with the corresponding preprocessing
        layer. The adapted preprocessing layers can be accessed from the adapted_preprocessors attribute.

        Args:
            dataset (tf.data.Dataset): Training data.
        """
        for _, preprocessor, features in self.feature_preprocessor_list:
            # get initial preprocessing layer config
            config = preprocessor.get_config()
            for k, feature in enumerate(features):
                # get a fresh preprocessing instance
                cloned_preprocessor = preprocessor if k==0 else preprocessor.from_config(config)
                feature_ds = extract_feature_column_tensorflow(dataset, feature)
                # check if layer has adapt method
                cloned_preprocessor.adapt(feature_ds, *args, **kwargs)
                self.adapted_preprocessors[feature] = cloned_preprocessor

    # what if inputs is not of type dict?
    @tf.function
    def call(self, inputs):
        """Apply adapted layers on new data.

        Args:
            inputs (dict): Dictionary of Tensors.

        Returns:
            dict: Dict of Tensors
        """
        forward_pass = dict()
        for name, _, features in self.feature_preprocessor_list:
            forward_pass[name] = [
                self.adapted_preprocessors[feature](inputs[feature])
                for feature in features
            ]
        return forward_pass

    @classmethod
    def from_infered_pipeline(cls, dataset):
        """Infer preprocessing spec from tf.data.Dataset type

        Args:
            dataset (tf.data.Dataset): Training data that contains features and/or target

        Returns:
            list: steps for preprocessing list
        """
        numeric_features = []
        categoric_features = []
        string_categoric_features = []
        for feature, _type in dataset.element_spec[0].items():
            if _type.dtype == tf.string:
                string_categoric_features.append(feature)
            elif _type.dtype == tf.int64:
                categoric_features.append(feature)
            else:
                numeric_features.append(feature)

        feature_preprocessor_list = [
            # FIXME: We will most likely not have all these steps
            (
                "numerical_features", layers.Normalization(), numeric_features
            ),
            (
                "categorical_features", layers.IntegerLookup(output_mode="multi_hot"), categoric_features,
            ),
            (
                "string_categorical_features", StringToIntegerLookup(), string_categoric_features,
            ),
        ]
        return feature_preprocessor_list


class FeaturePreprocessorFromPandasDataFrame(tf.keras.layers.Layer, BaseFeaturePreprocessor):
    """Feature Preprocessing Layer for pandas DataFrame type. The class takes in steps of preprocessing that
    needs to be applied on the dataset. The adapt method records all stateful parameters for each of the steps
    and features that will be adapted. These adapted layers will be stored as an attribute and applied in the
    forward pass of the network.

    Args:
        feature_encoder_list : List of preprocessor of the form: ('name', preprocessor type, list of features)

    """

    def __init__(self, feature_preprocessor_list=[], *args, **kwargs):
        super(FeaturePreprocessorFromPandasDataFrame, self).__init__(*args, *kwargs)
        self.feature_preprocessor_list = self.map_preprocessor(feature_preprocessor_list)
        self.adapted_preprocessors = dict()

    def adapt(self, dataset, *args, **kwargs):
        """Adapt layers from tf.data.Dataset source type.

        Args:
            dataset (tf.data.Dataset): Training data.
        """
        for _, preprocessor, features in self.feature_preprocessor_list:
            # get initial preprocessing layer config
            config = preprocessor.get_config()
            for k, feature in enumerate(features):
                # get a fresh preprocessing instance
                cloned_preprocessor = preprocessor if k==0 else preprocessor.from_config(config)
                feature_ds = extract_feature_column_pandas(dataset, feature)
                # check if layer has adapt method
                cloned_preprocessor.adapt(feature_ds, *args, **kwargs)
                self.adapted_preprocessors[feature] = cloned_preprocessor

    @tf.function
    def call(self, inputs):
        """Apply adapted layers on new data

        Args:
            inputs (dict): Dictionary of Tensors.

        Returns:
            dict: Dict of Tensors
        """
        forward_pass = dict()
        for name, _, features in self.feature_preprocessor_list:
            forward_pass[name] = [
                self.adapted_preprocessors[feature](inputs[feature])
                for feature in features
            ]
        return forward_pass

    @classmethod
    def from_infered_pipeline(cls, dataset):
        """Infer preprocessing spec from pd.DataFrame type

        Args:
            dataset (pd.DataFrame): Training data that contains features and/or target

        Returns:
            list: steps for preprocessing list
        """
        numeric_features = []
        categoric_features = []
        string_categoric_features = []

        for feature, _type in dataset.dtypes.items():
            if str(_type) == 'object':
                string_categoric_features.append(feature)
            elif str(_type) in ['int64', 'int32']:
                categoric_features.append(feature)
            else:
                numeric_features.append(feature)

        feature_preprocessor_list = [
            # FIXME: We will most likely not have all these steps
            (
                "numerical_features", layers.Normalization(), numeric_features
            ),
            (
                "categorical_features", layers.IntegerLookup(output_mode="multi_hot"), categoric_features,
            ),
            (
                "string_categorical_features", StringToIntegerLookup(), string_categoric_features,
            ),
        ]
        return feature_preprocessor_list


def extract_feature_column_tensorflow(dataset, name):
    """Extract feature from dataset

    Args:
        dataset (tf.data.Dataset): Training Data
        name (str): Name of feature to extract

    Returns:
        tf.data.Dataset: Mapped dataset for supplied feature.
    """
    feature = dataset.map(lambda x, y: x[name])
    feature = feature.map(lambda x: tf.expand_dims(x, -1))
    return feature


def extract_feature_column_pandas(dataset, name):
    """Extract feature from dataset

    Args:
        dataset (pd.DataFrame): Training Data
        name (str): Name of feature to extract

    Returns:
        array: numpy array of supplied feature.
    """
    return dataset[[name]].values



"""Preprocessing pipelines 
"""

from easyflow.preprocessing.pipeline import (
    FeaturePreprocessor,
    FeatureUnion,
)
from easyflow.preprocessing.custom import (
    FeatureInputLayer,
    NumericPreprocessingLayer,
    Pipeline, 
    PreprocessorChain,
    MultiOutputTransformer,
    StringToIntegerLookup
)


__all__ = [
    "FeatureInputLayer",
    "FeaturePreprocessor",
    "FeatureUnion",
    "NumericPreprocessingLayer",
    "Pipeline",
    "PreprocessorChain",
    "MultiOutputTransformer",
    "StringToIntegerLookup",
]



"""Testing for preprocessing pipeline module
"""

import unittest

import pandas as pd
import tensorflow as tf
from tensorflow.keras import layers

# local imports
from easyflow.data import TensorflowDataMapper
from easyflow.preprocessing import FeaturePreprocessor, FeatureUnion
from easyflow.preprocessing import (
    FeatureInputLayer,
    PreprocessorChain,
)


class TestPreprocessingPipelines(unittest.TestCase):
    """Test the Preprocessing module pipelines
    """

    def setUp(self):
        self.dataframe = pd.read_csv("easyflow/tests/test_data/heart.csv")
        labels = self.dataframe.pop("target")
        dataset_mapper = TensorflowDataMapper()
        self.dataset = dataset_mapper.map(self.dataframe, labels).batch(32)

        self.numerical_features = [
            "age",
            "trestbps",
            "chol",
            "thalach",
            "oldpeak",
            "slope",
        ]
        self.categorical_features = ["sex", "cp", "fbs", "restecg", "exang", "ca"]
        # thal is represented as a string
        self.string_categorical_features = ["thal"]
        self.dtype_mapper = {
            "age": tf.float32,
            "sex": tf.float32,
            "cp": tf.float32,
            "trestbps": tf.float32,
            "chol": tf.float32,
            "fbs": tf.float32,
            "restecg": tf.float32,
            "thalach": tf.float32,
            "exang": tf.float32,
            "oldpeak": tf.float32,
            "slope": tf.float32,
            "ca": tf.float32,
            "thal": tf.string,
        }
        self.all_feature_inputs = FeatureInputLayer(self.dtype_mapper)

        self.feature_preprocessor_list = [
            ("numeric_encoder", layers.Normalization(), self.numerical_features),
            (
                "categorical_encoder",
                layers.IntegerLookup(output_mode="multi_hot"),
                self.categorical_features,
            ),
            # For feature thal we first need to run StringLookup followed by a IntegerLookup layer
            (
                "string_encoder",
                PreprocessorChain(
                    [layers.StringLookup(), layers.IntegerLookup(output_mode="multi_hot")]
                ),
                self.string_categorical_features,
            ),
        ]

    def test_feature_preprocessor(self):
        """Test FeaturePreprocessor class
        """
        preprocessor = FeaturePreprocessor(self.feature_preprocessor_list)
        preprocessor.adapt(self.dataframe)
        preprocessing_layer = preprocessor(self.all_feature_inputs)
        assert list(preprocessing_layer.keys()) == ['numeric_encoder', 'categorical_encoder', 'string_encoder']
        assert preprocessing_layer.numeric_encoder.shape[-1] == 6

    def test_preprocessing_pipeline(self):
        """Test Feature union and model fit
        """
        all_feature_inputs, history = train_model_util(
            self.dataset, self.all_feature_inputs, self.feature_preprocessor_list
        )
        assert len(all_feature_inputs) == 13
        # test if the model ran through all 10 epochs
        assert len(history.history["loss"]) == 10

    def test_identity_layer(self):
        """Test that if preprocessor is None, the IdentityPreprocessing layer should be applied
        """
        feature_encoder_list_none = [
            ("numeric_encoder", None, self.numerical_features),
            (
                "categorical_encoder",
                layers.IntegerLookup(output_mode="multi_hot"),
                self.categorical_features,
            ),
            # For feature thal we first need to run StringLookup followed by a IntegerLookup layer
            (
                "string_encoder",
                PreprocessorChain(
                    [layers.StringLookup(), layers.IntegerLookup(output_mode="multi_hot")]
                ),
                self.string_categorical_features,
            ),
        ]

        all_feature_inputs, history = train_model_util(
            self.dataset, self.all_feature_inputs, feature_encoder_list_none
        )
        assert len(all_feature_inputs) == 13
        # test if the model ran through all 10 epochs
        assert len(history.history["loss"]) == 10

    def test_preprocessing_layer_with_no_args(self):
        """Test preprocessing layers with no arguments supplied
        """
        steps_list = [
            (
                "string_encoder",
                PreprocessorChain(
                    [layers.StringLookup(), layers.IntegerLookup(output_mode="multi_hot")]
                ),
                self.string_categorical_features,
            )
        ]
        preprocessor = FeatureUnion(steps_list)
        preprocessor.adapt(self.dataset)
        preprocessing_layer = preprocessor(self.all_feature_inputs)
        assert preprocessing_layer.shape[-1] == 6

    def test_encoding_validator(self):
        """Test preprocessing layers with arguments supplied
        """
        steps_list = [
            (
                "string_encoder",
                PreprocessorChain(
                    [
                        layers.StringLookup(max_tokens=4),
                        layers.IntegerLookup(output_mode="multi_hot"),
                    ]
                ),
                self.string_categorical_features,
            )
        ]
        preprocessor = FeatureUnion(steps_list)
        preprocessor.adapt(self.dataset)
        preprocessing_layer = preprocessor(self.all_feature_inputs)
        assert preprocessing_layer.shape[-1] == 5

    def test_adapt_with_args(self):
        """Test pipeline with args passed to .adapt
        """
        steps_list = [
                ("numeric_encoder", layers.Normalization(), self.numerical_features)
        ]
        # Apply on tf.data dataset
        preprocessor = FeatureUnion(steps_list)
        preprocessor.adapt(self.dataset, steps=1)
        preprocessing_layer = preprocessor(self.all_feature_inputs)
        assert preprocessing_layer.shape[-1] == 6
        del preprocessor
        # Apply on dataframe
        preprocessor = FeatureUnion(steps_list)
        preprocessor.adapt(self.dataframe, steps=1)
        preprocessing_layer = preprocessor(self.all_feature_inputs)
        assert preprocessing_layer.shape[-1] == 6

    def test_infered_pipeline(self):
        """test infered pipeline"""
        preprocessor = FeatureUnion.from_infered_pipeline(self.dataset)
        preprocessor.adapt(self.dataset)
        preprocessing_layer = preprocessor(self.all_feature_inputs)
        # setup simple network
        x = tf.keras.layers.Dense(128, activation="relu")(preprocessing_layer)
        x = tf.keras.layers.Dropout(0.5)(x)
        outputs = tf.keras.layers.Dense(1, activation="sigmoid")(x)
        model = tf.keras.Model(inputs=self.all_feature_inputs, outputs=outputs)
        model.compile(
            optimizer=tf.keras.optimizers.Adam(),
            loss=tf.keras.losses.BinaryCrossentropy(),
            metrics=[
                tf.keras.metrics.BinaryAccuracy(name="accuracy"),
                tf.keras.metrics.AUC(name="auc"),
            ],
        )
        history = model.fit(self.dataset, epochs=10)
        assert len(self.all_feature_inputs) == 13
        # test if the model ran through all 10 epochs
        assert len(history.history["loss"]) == 10


def train_model_util(dataset, all_feature_inputs, feature_preprocessor_list):
    """help function for testing

    Args:
        dataset (tf.data.Dataset): Features Data to apply encoder on.
        feature_encoding_list (list): [List of encoders of the form: ('name', encoder type, list of features)
    """
    preprocessor = FeatureUnion(feature_preprocessor_list)
    preprocessor.adapt(dataset)
    preprocessing_layer = preprocessor(all_feature_inputs)

    # setup simple network
    x = tf.keras.layers.Dense(128, activation="relu")(preprocessing_layer)
    x = tf.keras.layers.Dropout(0.5)(x)
    outputs = tf.keras.layers.Dense(1, activation="sigmoid")(x)
    model = tf.keras.Model(inputs=all_feature_inputs, outputs=outputs)
    model.compile(
        optimizer=tf.keras.optimizers.Adam(),
        loss=tf.keras.losses.BinaryCrossentropy(),
        metrics=[
            tf.keras.metrics.BinaryAccuracy(name="accuracy"),
            tf.keras.metrics.AUC(name="auc"),
        ],
    )
    history = model.fit(dataset, epochs=10)
    return all_feature_inputs, history


if __name__ == "__main__":
    unittest.main()



"""Custom preprocessing layers"""

from tensorflow.python.keras.engine.base_preprocessing_layer import PreprocessingLayer
import tensorflow as tf
from tensorflow.keras import layers


class FeatureInputLayer:
    """Create model inputs for each Feature in the dataset of features that will be used for training. There are two
    options to create the input layer. By default we will be taking in a dict mapper with feature name and dtype as input to
    construct inputs to the network. Second option is to infer the dtype from Dataset's element spec if source data comes from
    tf.data.Dataset type.

    Args:
        data_type_mapper (dict): Dictionary with feature as key and dtype as value
                                For example {'age': tf.float32, ...}
    Returns:
        (dict): Keras Input for each feature
    """

    def __new__(cls, dtype_mapper):
        """Create Input Layer with type mapper by default and return dict of Input objects.

        Args:
            dtype_mapper (dict): Key and Value pairs for feature name as key and data type as value

        Returns:
            dict: Dict with feaure name as key and  Input object value
        """
        return cls.from_dict_mapper(dtype_mapper)

    @classmethod
    def from_dict_mapper(cls, dtype_mapper):
        """Create model inputs

        Args:
            data_type_mapper (dict): Dictionary with feature as key and dtype as value
                                    For example {'age': tf.float32, ...}
        Returns:
            (dict): Keras Input for each feature
        """
        return {
            feature: tf.keras.Input(shape=(1,), name=feature, dtype=dtype)
            for feature, dtype in dtype_mapper.items()
        }

    @classmethod
    def infer_from_data(cls, dataset):
        """Create model inputs from the dataset's element spec.

        Args:
            input (tf.data.Dataset): Training dataset with features

        Returns:
            (dict): Keras Input for each feature
        """

        def get_data_spec(ds):
            """helper function to get extract element spec based on Dataset spec"""
            if len(ds.element_spec) == 2:
                return ds.element_spec[0]
            return ds.element_spec

        spec = get_data_spec(dataset)

        return {
            feature: tf.keras.Input(shape=(1,), name=feature, dtype=tspec.dtype)
            for feature, tspec in spec.items()
        }


class NumericPreprocessingLayer(PreprocessingLayer):
    """Helper class to apply no preprocessing and use feature as is"""

    def call(self, inputs):
        return tf.keras.layers.Reshape((1,))(inputs)

    def get_config(self):
        """Override get_config to ensure saving of models works"""
        return super().get_config().copy()

    def update_state(self, data):
        return dict()


class Pipeline(tf.keras.layers.Layer):
    """Preprocessing layer that chains one or more layers in a sequential order by
    subclassinig Layer class.

    Args:
        layers_to_adapt (list): List of layer that needs to be adapted
    """

    def __init__(self, layers_to_adapt, **kwargs):
        super(Pipeline, self).__init__(**kwargs)
        if not isinstance(layers_to_adapt, (list, tuple)):
            layers_to_adapt = [layers_to_adapt]
        self.layers_to_adapt = layers_to_adapt
        self.adapted_layers = []
        self.pipeline = None

    def adapt(self, data, *args, **kwargs):
        """Adapt layers to adapt sequentially and store adapted layers in executed order in adapted layer list.

        Args:
            data (tf.data.Dataset): Mapped tf.data.Dataset containing only the single feature
        """
        for layer in self.layers_to_adapt:
            if hasattr(layer, 'adapt'):
                layer.adapt(data, *args, **kwargs)
            self.adapted_layers.append(layer)
            if len(self.layers_to_adapt) >= 2:
                data = (
                    data.map(layer)
                    if isinstance(data, tf.data.Dataset)
                    else layer(data)
                )

        self.pipeline = tf.keras.models.Sequential(self.adapted_layers)

    def call(self, inputs):
        """Apply sequential model containing adapted layers.

        Args:
            inputs (tf.data.Dataset): Feature to be adapted as a Mapped Dataset

        Returns:
            tf.Tensor: returns output after applying adapted layers.
        """
        return self.pipeline(inputs)

    def get_config(self):
        """Update config with layers_to_adapt attr

        Returns:
            dict: Updated config
        """
        config = super().get_config()
        config.update(
            {"layers_to_adapt": self.layers_to_adapt, "pipeline": self.pipeline}
        )
        return config


class PreprocessorChain(tf.keras.models.Sequential):
    """Preprocessing model that chains one or more layers in a sequential order by subclassing
    Sequential model class. The functionality is the same as Pipeline.
    """

    def __init__(self, layers_to_adapt=[], **kwargs):
        super(PreprocessorChain, self).__init__(layers=[], **kwargs)
        if not isinstance(layers_to_adapt, (list, tuple)):
            layers_to_adapt = [layers_to_adapt]
        self.layers_to_adapt = layers_to_adapt

    def adapt(self, data, *args, **kwargs):
        """Adapt layers to adapt sequentially and add layers with the tf.keras.models.Sequential add 
        method.

        Args:
            data (tf.data.Dataset): Mapped tf.data.Dataset containing only the single feature
        """
        for counter, layer in enumerate(self.layers_to_adapt):
            if hasattr(layer, 'adapt'):
                layer.adapt(data, *args, **kwargs)

            super().add(layer)

            if len(self.layers_to_adapt) > counter:
                data = (
                    data.map(layer)
                    if isinstance(data, tf.data.Dataset)
                    else layer(data)
                )

    def get_config(self):
        """Update config with layers_to_adapt attr

        Returns:
            dict: Updated config
        """
        config = super().get_config()
        config.update({"layers_to_adapt": self.layers_to_adapt})
        return config


class MultiOutputTransformer(tf.keras.layers.Layer):
    """Applies different feature transformations or preprocessing on the same feature and concatenates
    the transformations into a single layer containing output of N transformations.
    """
    # we should be able to use this class the same way we do FeaturePreprocessor class, i.e standalone
    def __init__(self, steps=[], **kwargs):
        super(MultiOutputTransformer, self).__init__(**kwargs)
        if not isinstance(steps, (list, tuple)):
            steps = [steps]
        self.steps = steps
        self.processed_layers = []

    def adapt(self, data, *args, **kwargs):
        """Apply different or independent preprocessing steps on the same data. The results will be concatenated
        into a single layer.

        Args:
            data (tf.data.Dataset): Mapped tf.data.Dataset containing only the single feature
        """
        for layer in self.steps:
            if hasattr(layer, 'adapt'):
                layer.adapt(data, *args, **kwargs)
            self.processed_layers.append(layer)

    def call(self, inputs):
        """Apply sequential model containing adapted layers.

        Args:
            inputs (tf.data.Dataset): Feature to be adapted as a Mapped Dataset

        Returns:
            tf.Tensor: returns output after applying adapted layers.
        """
        return tf.keras.layers.concatenate([
            processed_layer(inputs) for processed_layer in self.processed_layers
        ])

    def get_config(self):
        """Update config with layers_to_adapt attr

        Returns:
            dict: Updated config
        """
        config = super().get_config()
        config.update(
            {"steps": self.steps, "processed_layers": self.processed_layers}
        )
        return config


def StringToIntegerLookup(**kwargs):
    """Implements a common pipeline when categorical features are of type string.
    Steps involve to first apply StringLookup followed by IntegerLookup.

    Args:
        kwargs: All arguments are related to IntegerLookup
    """
    return PreprocessorChain(
            [layers.StringLookup(), layers.IntegerLookup(output_mode='multi_hot', **kwargs)]   
    )



"""
Interface to use tensorflow feature columns with Keras. Only most used encoders
"""
import tensorflow as tf

from .base import BaseCategoricalFeatureColumnEncoder, get_unique_vocabulary


class CategoricalFeatureEncoder(BaseCategoricalFeatureColumnEncoder):
    """
    Class encodes Categorical features using tensorflow feature_columns. This is a wrapper to
    tf.feature_column.indicator_column to conform to the BaseFeatureColumnEncoder interface and does not change the behaviour.

    Examples
    --------
    >>> data = {'feature': ['a', 'b', 'c', 'c']}
    >>> dataset=tf.data.Dataset.from_tensor_slices(data).batch(4)
    >>> example_batch = next(iter(dataset))
    >>> feature_layer = lambda feature_column, batch: tf.keras.layers.DenseFeatures(feature_column)(batch).numpy()
    >>> encoder = CategoricalFeatureEncoder()
    >>> encoded_feature = encoder.encode(dataset, ['feature'])
    >>> feature_layer(encoded_feature, example_batch)
        array([[1., 0., 0.],
               [0., 1., 0.],
               [0., 0., 1.],
               [0., 0., 1.]], dtype=float32)
    """
    def __init__(self, **kwargs):
        super().__init__(feature_transformer=tf.feature_column.indicator_column, **kwargs)


class EmbeddingFeatureEncoder(BaseCategoricalFeatureColumnEncoder):
    """
    Class encodes high cardinality Categorical features(Embeddings) using tensorflow feature_columns. This is a wrapper to
    tf.feature_column.embedding_column to conform to the BaseFeatureColumnEncoder interface and does not change the behaviour.

    Examples
    --------
    >>> data = data = {'feature': ['a', 'b', 'c', 'c', 'c', 'c', 'd', 'd', 'd']}
    >>> dataset=tf.data.Dataset.from_tensor_slices(data).batch(4)
    >>> example_batch = next(iter(dataset))
    >>> feature_layer = lambda feature_column, batch: tf.keras.layers.DenseFeatures(feature_column)(batch).numpy()
    >>> encoder = EmbeddingFeatureEncoder(dimension=2)
    >>> encoded_feature = encoder.encode(dataset, ['feature'])
    >>> feature_layer(encoded_feature, example_batch)
        # results will differ due the randomness
        array([[-0.5162831 , -0.18130358],
               [ 0.50860345, -0.00483216],
               [ 0.00722361,  0.14289041],
               [ 0.00722361,  0.14289041]], dtype=float32)
    """
    def __init__(self, **kwargs):
        super().__init__(feature_transformer=tf.feature_column.embedding_column, **kwargs)


class CategoryCrossingFeatureEncoder(BaseCategoricalFeatureColumnEncoder):
    """Class creates cross column features using tensorflow feature_columns. This is a wrapper to
    tf.feature_column.crossed_column to conform to the BaseFeatureColumnEncoder interface and does not change the behaviour.

    Examples
    --------
    >>> data = {'feature_a': ['a', 'b', 'c', 'c', 'c', 'c', 'd', 'd', 'd'],
                'feature_b': ['e', 'f', 'g', 'g', 'g', 'g', 'g', 'g', 'g']}
    >>> dataset=tf.data.Dataset.from_tensor_slices(data).batch(4)
    >>> example_batch = next(iter(dataset))
    >>> feature_layer = lambda feature_column, batch: tf.keras.layers.DenseFeatures(feature_column)(batch).numpy()
    >>> encoder = CategoryCrossingFeatureEncoder(hash_bucket_size=3)
    >>> encoded_feature = encoder.encode(dataset, ['feature_a', 'feature_b'])
    >>> feature_layer(encoded_feature, example_batch)
        array([[0., 0., 1.],
               [1., 0., 0.],
               [0., 0., 1.],
               [0., 0., 1.]], dtype=float32)
    """
    def __init__(self, **kwargs):
        super().__init__(feature_transformer=tf.feature_column.crossed_column, **kwargs)

    def encode(self, dataset=None, features=None):
        """Apply Cross column feature engineering followed by an indicator column

        Args:
            dataset (tf.data.Dataset): Features Data to apply encoder on.
            features (list): list of feature names

        Returns:
            (list): list of encoded features
        """
        feature_vocab_list = get_unique_vocabulary(dataset, features)
        crossed_features = self.feature_transformer(feature_vocab_list, **self.kwargs)
        crossed_features = tf.feature_column.indicator_column(crossed_features)
        return crossed_features


class EmbeddingCrossingFeatureEncoder(BaseCategoricalFeatureColumnEncoder):
    """Class creates cross column features using tensorflow feature_columns. This is a wrapper to
    tf.feature_column.embedding_column to conform to the BaseFeatureColumnEncoder interface and does not change the behaviour.

    Examples
    --------
    >>> data = {'feature_a': ['a', 'b', 'c', 'c', 'c', 'c', 'd', 'd', 'd'],
                'feature_b': ['e', 'f', 'g', 'g', 'g', 'g', 'g', 'g', 'g']}
    >>> dataset=tf.data.Dataset.from_tensor_slices(data).batch(4)
    >>> example_batch = next(iter(dataset))
    >>> feature_layer = lambda feature_column, batch: tf.keras.layers.DenseFeatures(feature_column)(batch).numpy()
    >>> encoder = EmbeddingCrossingFeatureEncoder(embedding_dimension=2, hash_bucket_size=6)
    >>> encoded_feature = encoder.encode(dataset, ['feature_a', 'feature_b'])
    >>> feature_layer(encoded_feature, example_batch)
        array([[-0.49085316,  0.18231249],
               [ 0.05757887,  0.56533635],
               [-0.49085316,  0.18231249],
               [-0.49085316,  0.18231249]], dtype=float32)
    """
    def __init__(self, embedding_dimension, **kwargs):
        super().__init__(feature_transformer=tf.feature_column.crossed_column, **kwargs)
        self.embedding_dimension = embedding_dimension

    def encode(self, dataset=None, features=None):
        """Apply Cross column feature engineering followed by an embedding column

        Args:
            dataset (tf.data.Dataset): Features Data to apply encoder on.
            features (list): list of feature names

        Returns:
            (list): list of encoded features
        """
        feature_vocab_list = get_unique_vocabulary(dataset, features)
        crossed_features = self.feature_transformer(feature_vocab_list, **self.kwargs)
        embedded_crossed_features = tf.feature_column.embedding_column(categorical_column=crossed_features,
                                                                        dimension=self.embedding_dimension)
        return embedded_crossed_features



"""
Interface to use tensorflow feature columns with Keras for Numerical features
"""

import tensorflow as tf

from .base import BaseFeatureColumnEncoder


class NumericalFeatureEncoder(BaseFeatureColumnEncoder):
    """
    Class encodes numerical features using tensorflow feature_columns. This is a wrapper to
    tf.feature_column.numerical_column to conform to the BaseFeatureColumnEncoder interface and does not change the behaviour.

    Examples
    --------
    >>> data = {'feature': [1.1, 1.2, 0.0, 2.2]}
    >>> dataset = tf.data.Dataset.from_tensor_slices(data).batch(4)
    >>> example_batch = next(iter(dataset))
    >>> feature_layer = lambda feature_column, batch: tf.keras.layers.DenseFeatures(feature_column)(batch).numpy()
    >>> encoder = NumericalFeatureEncoder()
    >>> encoded_feature = encoder.encode(dataset, ['feature'])
    >>> feature_layer(encoded_feature, example_batch)
        array([[1.1],
               [1.2],
               [0. ],
               [2.2]], dtype=float32)
    """
    def __init__(self, **kwargs):
        super().__init__(feature_transformer=tf.feature_column.numeric_column, **kwargs)


class BucketizedFeatureEncoder(BaseFeatureColumnEncoder):
    """Class applies bucketized encoding by first applying NumericalFeatureEncoder. This is a wrapper to
    tf.feature_column.bucketized_column to conform to the BaseFeatureColumnEncoder interface and does not change the behaviour.

    Examples
    --------
    >>> data = {'feature': [1.1, 1.2, 0.0, 2.2]}
    >>> dataset = tf.data.Dataset.from_tensor_slices(data).batch(4)
    >>> example_batch = next(iter(dataset))
    >>> feature_layer = lambda feature_column, batch: tf.keras.layers.DenseFeatures(feature_column)(batch).numpy()
    >>> encoder = BucketizedFeatureEncoder(boundaries=[1])
    >>> encoded_feature = encoder.encode(dataset, ['feature'])
    >>> feature_layer(encoded_feature, example_batch)
        array([[0., 1.],
               [0., 1.],
               [1., 0.],
               [0., 1.]], dtype=float32)
    """
    def __init__(self, **kwargs):
        super().__init__(feature_transformer=tf.feature_column.bucketized_column, **kwargs)

    def encode(self, dataset=None, features=None):
        """Apply bucketized feature encoding by first applying numerical encoder.

        Args:
            dataset (tf.data.Dataset): Features Data to apply encoder on.
            features (list): list of feature names

        Returns:
            (list): list of encoded features
        """
        numerical_encoder = NumericalFeatureEncoder()
        numerical_inputs, numerical_encoded_features = numerical_encoder.transform(dataset, features)
        return [self.feature_transformer(feature, **self.kwargs) for feature in numerical_encoded_features]



"""base class and usefull methods"""
from easyflow.preprocessing.custom import NumericPreprocessingLayer


class BaseFeaturePreprocessor:
    """Interface for all Feature Preprocessor
    """
    def map_preprocessor(self, steps):
        """Check and Map input if any of the preprocessors are None, i.e. use as is. For 
        example Binary features that don't need further preprocessing   

        Args:
            steps (list): Preprocessing steps.

        Returns:
            list: List of mapped preprocessors if None was supplied.
        """
        selector = lambda _preprocessor: _preprocessor or NumericPreprocessingLayer()
        return [
            (name, selector(preprocessor), step) for name, preprocessor, step in steps
        ]

    def __getitem__(self, idx):
        # This should rather return the adapted layers for the specific step
        return self.feature_preprocessor_list[idx]

    def __len__(self):
        """Total number of steps

        Returns:
            int: Total number of steps
        """
        return len(self.feature_preprocessor_list)

    @property
    def preprocessor_name(self):
        """Return the step names

        Returns:
            list: List of step names
        """
        return [self[k][0] for k in range(len(self))]

    @classmethod
    def from_infered_pipeline(cls, dataset):
        """Infer standard pipeline for structured data, i.e Normalization for numerical
        features and StringLookup/IntegerLookup for categoric features

        Args:
            dataset (tf.data.Dataset): Features Data to apply encoder on.

        Returns:
            BaseFeaturePreprocessorLayer: Initilized BaseFeaturePreprocessorLayer object
        """
        raise NotImplementedError



"""Keras and tensorflow feature column pipeline
"""

from .base import BaseFeatureColumnEncoder, BaseCategoricalFeatureColumnEncoder, get_unique_vocabulary
from .categorical_encoders import CategoricalFeatureEncoder, EmbeddingFeatureEncoder,\
    CategoryCrossingFeatureEncoder, EmbeddingCrossingFeatureEncoder
from .numerical_encoders import NumericalFeatureEncoder, BucketizedFeatureEncoder
from .pipeline import FeatureColumnTransformer, FeatureUnionTransformer


__all__ = ["BaseFeatureColumnEncoder",
           "BaseCategoricalFeatureColumnEncoder",
           "CategoricalFeatureEncoder",
           "EmbeddingFeatureEncoder",
           "CategoryCrossingFeatureEncoder",
           "EmbeddingCrossingFeatureEncoder",
           "NumericalFeatureEncoder",
           "BucketizedFeatureEncoder",
           "get_unique_vocab",
           "FeatureColumnTransformer",
           "FeatureUnionTransformer"]



import pathlib
import io
import os
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text(encoding='utf-8')
REQUIREMENTS = (pathlib.Path(__file__).parent / "requirements.txt").read_text().splitlines()[1:]
REQUIRES_PYTHON = '>=3.7.0'


setup(
    name="easy-tensorflow",
    version="1.4.1",
    author="Fernando Nieuwveldt",
    author_email="fdnieuwveldt@gmail.com",
    description="Feature Pipelines for Keras preprocessing layers.",
    long_description_content_type='text/markdown',
    long_description=README,
    url="https://github.com/fernandonieuwveldt/easyflow",
    python_requires=REQUIRES_PYTHON,
    packages=find_packages(exclude='tests'),
    install_requires=REQUIREMENTS,
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



"""
Feature transforming pipelines/composers using tensorflow feature columns
"""

import tensorflow as tf

from easyflow.feature_encoders import NumericalFeatureEncoder, CategoricalFeatureEncoder


class FeatureColumnTransformer:
    """Apply column based transformation on the data

    Examples
    --------
    >>> data = {'feature_a': ['a', 'b', 'c', 'c'],
                'feature_b': [1.1, 1.2, 0.0, 2.2]}
    >>> target = {'target': [1, 1, 0, 0]}
    >>> dataset=tf.data.Dataset.from_tensor_slices((data, target)).batch(4)
    >>> example_batch = next(iter(dataset))
    >>> feature_encoder_list = [('categorical_encoder', CategoricalFeatureEncoder(), ['feature_a']),
                                ('bucketized_encoder', BucketizedFeatureEncoder(boundaries=[1]), ['feature_b'])]
    >>> encoder = FeatureColumnTransformer(feature_encoder_list)
    >>> inputs, feature_layer = encoder.transform(dataset)

    Args:
        feature_encoder_list : List of encoders of the form: ('name', encoder type, list of features)
    """
    def __init__(self, feature_encoder_list=None):
        self.feature_encoder_list = feature_encoder_list

    @classmethod
    def infer_feature_transformer(cls, dataset):
        """Infer standard pipeline for structured data, i.e NumericalFeatureEncoder for numeric
        features and CategoricalFeatureEncoder for categoric features

        Args:
            dataset (tf.data.Dataset): Features Data to apply encoder on.

        Returns:
            (list): basic encoding list
        """
        numeric_features = []
        categoric_features = []
        for feature, _type in dataset.element_spec[0].items():
            if _type.dtype in [tf.string, tf.int64]:
                categoric_features.append(feature)
            else:
                numeric_features.append(feature)
        feature_encoder_list = [('numerical_features', NumericalFeatureEncoder(), numeric_features),
                                ('categorical_features', CategoricalFeatureEncoder(), categoric_features)]
        return cls(feature_encoder_list)

    def transform(self, dataset):
        """Apply feature encodings on supplied list

        Args:
            dataset (tf.data.Dataset): Features Data to apply encoder on.

        Returns:
            (dict, list): Keras inputs for each feature and list of encoders
        """
        feature_layer_inputs = {}
        feature_encoder_layer = {}
        for (name, encoder, features) in self.feature_encoder_list:
            # best to use create_inputs and encode independently?
            feature_inputs, feature_encoded = encoder.transform(dataset, features)
            feature_layer_inputs.update(feature_inputs)
            feature_encoder_layer[name] = feature_encoded
        # Apply DenseFeatures layer when all feature layer inputs are available
        for name, encoded_feature in feature_encoder_layer.items():
            feature_encoder_layer[name] = tf.keras.layers.DenseFeatures(encoded_feature)(feature_layer_inputs)
        return feature_layer_inputs, feature_encoder_layer


class FeatureUnionTransformer(FeatureColumnTransformer):
    """Apply transformations and apply union/concatenating individual feature layers

    Examples
    --------
    >>> data = {'feature_a': ['a', 'b', 'c', 'c'],
                'feature_b': [1.1, 1.2, 0.0, 2.2]}
    >>> target = {'target': [1, 1, 0, 0]}
    >>> dataset=tf.data.Dataset.from_tensor_slices((data, target)).batch(4)
    >>> example_batch = next(iter(dataset))
    >>> feature_encoder_list = [('categorical_encoder', CategoricalFeatureEncoder(), ['feature_a']),
                                ('bucketized_encoder', BucketizedFeatureEncoder(boundaries=[1]), ['feature_b'])]
    >>> encoder = FeatureUnionTransformer(feature_encoder_list)
    >>> inputs, feature_layer = encoder.transform(dataset)

    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def transform(self, dataset):
        """Join features. If more flexibility and customization is needed use FeatureColumnTransformer.

        Args:
            dataset (tf.data.Dataset): Features Data to apply encoder on.

        Returns:
            (dict, list): Keras inputs for each feature and list of encoders
        """
        feature_layer_inputs, feature_encoder_layer = super().transform(dataset)
        feature_layer = [feature_layer for feature_layer in feature_encoder_layer.values()]
        if len(feature_layer) == 1:
            return feature_layer_inputs, feature_layer.pop()
        return feature_layer_inputs, tf.keras.layers.concatenate(feature_layer)



"""This class will serve as a factory for different data types"""

import pandas as pd
import tensorflow as tf
from easyflow.preprocessing.feature_preprocessor_layer import (
    FeaturePreprocessorFromTensorflowDataset,
    FeaturePreprocessorFromPandasDataFrame,
)


class FeaturePreprocessorFactory(tf.keras.models.Model):
    """Apply column based transformation on the data using tf.keras  preprocessing layers.

    Args:
        feature_encoder_list : List of preprocessor of the form: ('name', preprocessor type, list of features)
    """

    def __init__(self, feature_preprocessor_list=[], *args, **kwargs):
        super(FeaturePreprocessorFactory, self).__init__(*args, *kwargs)
        self.feature_preprocessor_list = feature_preprocessor_list
        self.preprocessor_flow = None

    @classmethod
    def from_infered_pipeline(cls, dataset):
        """Infer standard pipeline for structured data, i.e Normalization for numerical
        features and StringLookup/IntegerLookup for categoric features

        Args:
            dataset (tf.data.Dataset): Features Data to apply encoder on.

        Returns:
            FeaturePreprocessorFactory: Initilized FeaturePreprocessorFactory object
        """
        if isinstance(dataset, tf.data.Dataset):
            feature_preprocessor_list = (
                FeaturePreprocessorFromTensorflowDataset.from_infered_pipeline(dataset)
            )
            return cls(feature_preprocessor_list)

    def adapt(self, dataset, *args, **kwargs):
        """Adapt preprocessing layers.

        Args:
            dataset ([pandas.DataFrame, tf.data.Dataset]): Training data.
        """
        if isinstance(dataset, tf.data.Dataset):
            self.preprocessor_flow = FeaturePreprocessorFromTensorflowDataset(
                self.feature_preprocessor_list
            )

        if isinstance(dataset, pd.DataFrame):
            self.preprocessor_flow = FeaturePreprocessorFromPandasDataFrame(
                self.feature_preprocessor_list
            )

        if not self.preprocessor_flow:
            raise Exception("Datatype not supported: dataset should be of type pandas DataFrame or Tensorflow tf.data.Dataset")

        self.preprocessor_flow.adapt(dataset, *args, **kwargs)

    def call(self, inputs):
        """Apply adapted layers on new data

        Args:
            inputs (dict): Dictionary of Tensors.

        Returns:
            dict: Dict of Tensors
        """
        if not self.preprocessor_flow:
            raise Exception("First run adapt before forward pass.")

        return self.preprocessor_flow(inputs)



"""Base Classes for encoders using tensrflow feature columns"""

from abc import ABC, abstractmethod

import tensorflow as tf


class BaseFeatureColumnEncoder(ABC):
    """Base class for a tensorlow feature column based encoder"""

    def __init__(self, feature_transformer=None, **kwargs):
        self.feature_transformer = feature_transformer
        self.kwargs = kwargs

    def create_inputs(self, dataset=None, features=None):
        """Create model inputs

        Args:
            dataset (tf.data.Dataset): Features Data to apply encoder on.
            features (list): list of feature names

        Returns:
            (dict): Keras inputs for each feature
        """
        def get_data_type(ds, feature):
            """helper function to get dtype of feature"""
            if len(ds.element_spec) == 2:
                return ds._structure[0][feature].dtype
            return ds._structure[feature].dtype

        return {feature: tf.keras.Input(shape=(1,), name=feature, dtype=get_data_type(dataset, feature))\
            for feature in features}

    # should perhaps be a abstractmethod
    def encode(self, dataset=None, features=None):
        """Apply feature encoding. This method can be over ridden for specific use cases

        Args:
            dataset (tf.data.Dataset): Features Data to apply encoder on.
            features (list): list of feature names

        Returns:
            (list): list of encoded features
        """
        return [self.feature_transformer(feature, **self.kwargs) for feature in features]

    def transform(self, dataset=None, features=None):
        """Encoding numerical type features with tensorflow feature columns

        Args:
            dataset (tf.data.Dataset): Features Data to apply encoder on.
            features (list): list of feature names

        Returns:
            (dict, list): Keras inputs for each feature and list of encoders
        """
        return self.create_inputs(dataset, features), self.encode(dataset, features)


def get_unique_vocabulary(dataset=None, features=None):
    """Get feature vocabulary list

    Args:
        dataset (tf.data.Dataset): Features Data to apply encoder on.
        features (list): list of feature names

    Returns:
        (dict): dictionary containing list of unique vocabulary values for each feature
    """
    if hasattr(dataset, '_batch_size'):
        # unbatch dataset
        dataset = dataset.unbatch()

    if isinstance(dataset.element_spec, tuple):
        map_func = lambda x, y: x[feature]
    else:
        map_func = lambda x: x[feature]

    feature_vocab_list = {}
    for feature in features:
        uniq_vocab = list(dataset.map(map_func)
                                 .unique()
                                 .as_numpy_iterator())
        if all(map(lambda x: isinstance(x, bytes), uniq_vocab)):
            # map bytes to ensure objects are serializable when saving model
            uniq_vocab = [str(value, 'utf-8') for value in uniq_vocab]
        feature_vocab_list[feature] = tf.feature_column.categorical_column_with_vocabulary_list(feature, uniq_vocab)
    return feature_vocab_list


class BaseCategoricalFeatureColumnEncoder(BaseFeatureColumnEncoder):
    """Base class for categorical type encoders
    """
    def __init__(self, feature_transformer=None, **kwargs):
        super().__init__(feature_transformer=feature_transformer, **kwargs)

    def encode(self, dataset=None, features=None):
        """Apply feature encoding that requires unique vocabulary as input

        Args:
            dataset (tf.data.Dataset): Features Data to apply encoder on.
            features (list): list of feature names

        Returns:
            (list): list of encoded features
        """
        feature_vocab_list = get_unique_vocabulary(dataset, features)
        return [self.feature_transformer(feature_vocab_list[feature], **self.kwargs) for feature in features]
